// Copyright (C) 2015 Codership Oy <info@codership.com>

// $Id$

#ifndef __gcs_act_cchange_test__
#define __gcs_act_cchange_test__

#include <check.h>

Suite *gcs_act_cchange_suite(void);

#endif /* __gcs_act_cchange_test__ */
