//
// Copyright (C) 2015 Codership Oy <info@codership.com>
//

#ifndef GU_SHARED_PTR_TEST_HPP
#define GU_SHARED_PTR_TEST_HPP

#include <check.h>

extern Suite* gu_shared_ptr_suite(void);

#endif // ! GU_SHARED_PTR_TEST_HPP
